# -*- coding: utf-8 -*-

import base64
import os
from lamvery.actions.base import BaseAction
from termcolor import cprint

class InvokeAction(BaseAction):

    def __init__(self, args):
        super(InvokeAction, self).__init__(args)

        if os.path.exists(args.json):
            self._json = open(args.json, 'r').read()
        else:
            self._json = args.json
        self._version = args.version
        
        if args.alias is None:
            self._alias = self._config.get_default_alias()
        else:
            self._alias = args.alias

    def action(self):
        qualifier = self._alias
        client = self.get_client()

        if self._version is not None:
            qualifier = self._version

        ret = client.invoke(
            name=self._config.get_function_name(),
            qualifier=qualifier,
            payload=self._json)

        if ret.get('FunctionError') is None:
            cprint(base64.b64decode(ret.get('LogResult')), 'green')
        else:
            self._logger.error('{} error occurred'.format(ret.get('FunctionError')))
            cprint(base64.b64decode(ret.get('LogResult')), 'red')



# -*- coding: utf-8 -*-

from lamvery.actions.base import BaseAction
from lamvery.archive import Archive

class ArchiveAction(BaseAction):

    def __init__(self, args):
        super(ArchiveAction, self).__init__(args)
        self._single_file = args.single_file
        self._no_libs = args.no_libs

    def action(self):
        archive_name = self._config.get_archive_name()
        function_filename = self._config.get_function_filename()
        secret = self._config.generate_lambda_secret()
        exclude = self._config.get_exclude()
        archive = Archive(filename=archive_name,
                          function_filename=function_filename,
                          secret=secret,
                          single_file=self._single_file,
                          no_libs=self._no_libs,
                          exclude=exclude)
        zipfile = archive.create_zipfile()
        with open(archive_name, 'w') as f:
            f.write(zipfile.read())
        zipfile.close()
        self._logger.info('Output archive(zip) to {}'.format(archive_name))



# -*- coding: utf-8 -*-

from lamvery.actions.base import BaseAction

CONF_DIFF_KEYS = [
    ('Runtime', 'runtime',),
    ('Role', 'role',),
    ('Handler', 'handler',),
    ('Description', 'description',),
    ('Timeout', 'timeout',),
    ('MemorySize', 'memory_size',),
]

class ConfigureAction(BaseAction):

    def __init__(self, args):
        super(ConfigureAction, self).__init__(args)

    def action(self):
        func_name   = self._config.get_function_name()
        local_conf  = self._config.get_configuration()
        client      = self.get_client()
        remote_conf = client.get_function_conf(func_name)

        if len(remote_conf) > 0:
            self._print_diff(
                prefix='[Function]',
                remote=remote_conf, local=local_conf, keys=CONF_DIFF_KEYS)
            client.update_function_conf(local_conf)
        else:
            msg = '"{}" function is not exists. Please `deploy` at first.'.format(func_name)
            raise Exception(msg)



# -*- coding: utf-8 -*-

import os
import sys
import tempfile
import shutil
import re
import yaml
import lamvery.secret
import warnings
from zipfile import PyZipFile, ZIP_DEFLATED
from lamvery.log import get_logger

warnings.simplefilter("ignore", UserWarning)

EXCLUDE_DIR = [
    '.git',
    '__pycache__'
]

PYFILE_PATTERN = re.compile('.+\.py.?$')

class Archive:

    def __init__(self, filename, function_filename=None, single_file=False, no_libs=False, secret={}, exclude=[]):
        self._filename = filename
        self._function_filename = function_filename
        self._tmpdir = tempfile.mkdtemp(suffix='lamvery')
        self._zippath = os.path.join(self._tmpdir, self._filename)
        self._secretpath = os.path.join(self._tmpdir, lamvery.secret.SECRET_FILE_NAME)
        self._secret = secret
        self._single_file = single_file
        self._no_libs = no_libs
        self._exclude = exclude

    def __del__(self):
        shutil.rmtree(self._tmpdir)

    def create_zipfile(self):
        lamvery.secret.generate(self._secretpath, self._secret)
        with PyZipFile(self._zippath, 'w', compression=ZIP_DEFLATED) as zipfile:
            for p in self._get_paths():
                if os.path.isdir(p):
                    self._archive_dir(zipfile, p)
                else:
                    self._archive_file(zipfile, p)
            if not self._single_file:
                self._archive_file(zipfile, self._secretpath)
        return open(self._zippath, 'rb')

    def _archive_dir(self, zipfile, path):
        dirname = os.path.basename(path)
        if not self.is_exclude_dir(dirname):
            zipfile.writepy(path)
            for dirpath, dirnames, filenames in os.walk(path):
                for filename in filenames:
                    f_path = os.path.join(dirpath, filename)
                    if not self.is_source_file(filename) and not self.is_exclude_file(filename):
                        f_name = f_path.replace(os.path.dirname(path) + os.sep, '')
                        try:
                            zipfile.getinfo(f_name)
                        except KeyError:
                            zipfile.write(f_path, f_name)

    def _archive_file(self, zipfile, path):
        filename = os.path.basename(path)
        if not self.is_exclude_file(filename):
            if PYFILE_PATTERN.match(filename) is None:
                try:
                    zipfile.getinfo(filename)
                except KeyError:
                    zipfile.write(path, filename)
            elif filename.endswith('.py'):
                try:
                    zipfile.getinfo('{}c'.format(filename))
                    zipfile.getinfo('{}o'.format(filename))
                except KeyError:
                    if self._single_file:
                        zipfile.write(path, filename)
                    else:
                        zipfile.writepy(path)

    def is_exclude(self, name):
        for ex in self._exclude:
            if re.compile(ex).match(name) is not None:
                return True
        return False

    def is_exclude_file(self, name):
        if self.is_exclude(name):
            return True
        if name == self._filename:
            return True
        return False

    def is_exclude_dir(self, name):
        if self.is_exclude(name):
            return True
        for ex in EXCLUDE_DIR:
            if name == ex:
                return True
        return False

    def is_source_file(self, name):
        return PYFILE_PATTERN.match(name) is not None

    def _get_paths(self):
        logger = get_logger(__name__)
        try:
            venv = os.environ['VIRTUAL_ENV']
        except:
            logger.warn(
                'VIRTUAL_ENV environment variable can not be found. Python libraries are not included in the archive.')
            venv = None
        paths = []
        if not self._no_libs and venv is not None:
            for p in sys.path:
                if os.path.isdir(p) and os.path.exists(p):
                    if p.startswith(venv) and p.find('site-packages') != -1:
                        for f in os.listdir(p):
                            f_path = os.path.join(p ,f)
                            paths.append(f_path)
        for f in os.listdir(os.getcwd()):
            f_path = os.path.join(os.getcwd() ,f)
            if not f_path == venv:
                paths.append(f_path)
        if self._single_file:
            f = self._function_filename
            paths = [os.path.join(os.getcwd(), f)]
        return paths

    def get_size(self):
        return os.stat(self._zippath).st_size



from lamvery.actions.archive import ArchiveAction
from lamvery.actions.configure import ConfigureAction
from lamvery.actions.decrypt import DecryptAction
from lamvery.actions.deploy import DeployAction
from lamvery.actions.encrypt import EncryptAction
from lamvery.actions.events import EventsAction
from lamvery.actions.init import InitAction
from lamvery.actions.set_alias import SetAliasAction
from lamvery.actions.invoke import InvokeAction
from lamvery.actions.rollback import RollbackAction



# -*- coding: utf-8 -*-

import boto3
import botocore
import base64
import hashlib
import json

from lamvery.utils import previous_alias

class Client:

    def __init__(self, region=None, profile=None, dry_run=False):
        session = boto3.session.Session(
            profile_name=profile, region_name=region)
        self._lambda = session.client('lambda')
        self._kms = session.client('kms')
        self._events = session.client('events')
        self._dry_run = dry_run

    def get_function_conf(self, name):
        try:
            res = self._lambda.get_function(
                FunctionName=name)
            return res['Configuration']
        except botocore.exceptions.ClientError:
            return {}

    def create_function(self, zipfile, conf, publish):
        if not self._dry_run:
            self._lambda.create_function(
                FunctionName=conf['name'],
                Runtime=conf['runtime'],
                Role=conf['role'],
                Handler=conf['handler'],
                Code={
                    'ZipFile': zipfile.read(),
                },
                Description=conf['description'],
                Timeout=conf['timeout'],
                MemorySize=conf['memory_size'],
                Publish=publish,
            )

    def update_function_code(self, zipfile, conf, publish):
        if not self._dry_run:
            ret = self._lambda.update_function_code(
                FunctionName=conf['name'],
                ZipFile=zipfile.read(),
                Publish=publish)
            return ret['Version']
        return None

    def update_function_conf(self, conf):
        if not self._dry_run:
            self._lambda.update_function_configuration(
                FunctionName=conf['name'],
                Role=conf['role'],
                Handler=conf['handler'],
                Description=conf['description'],
                Timeout=conf['timeout'],
                MemorySize=conf['memory_size'])

    def get_alias(self, function, alias):
        try:
            return self._lambda.get_alias(
                FunctionName=function,
                Name=alias)
        except botocore.exceptions.ClientError:
            return {}

    def create_alias(self, function, alias, version):
        if not self._dry_run:
            self._lambda.create_alias(
                    FunctionName=function,
                    Name=alias,
                    FunctionVersion=version)

    def update_alias(self, function, alias, version):
        if not self._dry_run:
            self._lambda.update_alias(
                    FunctionName=function,
                    Name=alias,
                    FunctionVersion=version)

    def encrypt(self, key_id, text):
        res = self._kms.encrypt(KeyId=key_id, Plaintext=text)
        return base64.b64encode(res.get('CiphertextBlob'))

    def decrypt(self, cipher_text):
        res = self._kms.decrypt(CiphertextBlob=base64.b64decode(cipher_text))
        return res.get('Plaintext')

    def calculate_capacity(self, next_marker=None):
        if next_marker:
            r = self._lambda.list_functions(MaxItems=500, Marker=next_marker)
        else:
            r = self._lambda.list_functions(MaxItems=500)

        size = sum(
            self._calculate_versions_capacity(
                f['FunctionName']) for f in r['Functions'])

        if 'NextMarker' in r:
            return size + self.calculate_capacity(next_marker=r['NextMarker'])
        else:
            return size

    def _calculate_versions_capacity(self, function_name, next_marker=None):
        if next_marker:
            r = self._lambda.list_versions_by_function(
                FunctionName=function_name, MaxItems=500, Marker=next_marker)
        else:
            r = self._lambda.list_versions_by_function(
                FunctionName=function_name, MaxItems=500)

        size = sum(f['CodeSize'] for f in r['Versions'])

        if 'NextMarker' in r:
            return size + self._calculate_versions_capacity(
                function_name=function_name, next_marker=r['NextMarker'])
        else:
            return size

    def get_rules_by_target(self, arn):
        names = self._get_rule_names_by_tagert(arn=arn)

        rules = []
        for name in names:
            rules.append(self._events.describe_rule(Name=name))

        return rules

    def _get_rule_names_by_tagert(self, arn, next_token=None):
        try:
            if next_token:
                names = self._events.list_rule_names_by_target(
                    TargetArn=arn, NextToken=next_token, Limit=100)
            else:
                names = self._events.list_rule_names_by_target(
                    TargetArn=arn, Limit=100)
        except botocore.exceptions.ClientError:
            return []

        ret = names['RuleNames']
        if 'NextToken' in names:
            ret.extend(
                self._get_rule_names_by_tagert(
                    arn=arn, next_token=names['NextToken']))
        return ret

    def put_rule(self, rule):
        if self._dry_run:
            return {}
        else:
            kwargs = {}
            kwargs['Name'] = rule['rule']
            kwargs['State'] = rule.get('state', 'ENABLED')

            if rule.get('description') is not None:
                kwargs['Description'] = rule.get('description')
            if rule.get('pattern') is not None:
                kwargs['EventPattern'] = rule.get('pattern')
            if rule.get('schedule') is not None:
                kwargs['ScheduleExpression'] = rule.get('schedule')

            return self._events.put_rule(**kwargs)

    def put_targets(self, rule, targets, arn):
        if not self._dry_run:
            t_args = []

            for t in targets:
                arg = {'Id': t['id'], 'Arn': arn}
                if t.get('input') is not None:
                    arg['Input'] = t['input']
                if t.get('input_path') is not None:
                    arg['InputPath'] = t['input_path']
                t_args.append(arg)

            self._events.put_targets(Rule=rule, Targets=t_args)

    def get_targets_by_rule(self, rule, next_token=None):
        try:
            if next_token:
                rules = self._events.list_targets_by_rule(
                    Rule=rule, NextToken=next_token, Limit=100)
            else:
                rules = self._events.list_targets_by_rule(
                    Rule=rule, Limit=100)
        except botocore.exceptions.ClientError:
            return []

        ret = rules['Targets']
        if 'NextToken' in rules:
            ret.extend(
                self.get_targets_by_rule(
                    rule=rule, next_token=rules['NextToken']))
        return ret

    def remove_targets(self, rule, targets):
        if not self._dry_run:
            self._events.remove_targets(Rule=rule, Ids=targets)

    def delete_rule(self, name):
        if not self._dry_run:
            self._events.delete_rule(Name=name)

    def add_permission(self, function, rule_name, rule_arn):
        if not self._dry_run:
            try:
                self._lambda.add_permission(
                    FunctionName=function,
                    Action='lambda:InvokeFunction',
                    Principal='events.amazonaws.com',
                    SourceArn=rule_arn,
                    StatementId=self._generate_statement_id(function, rule_name))
            except botocore.exceptions.ClientError:
                pass

    def remove_permission(self, function, rule):
        if not self._dry_run:
            self._lambda.remove_permission(
                FunctionName=function,
                StatementId=self._generate_statement_id(function, rule))

    def _generate_statement_id(self, function, rule):
        return hashlib.sha256(
            'lamvery-{}-{}'.format(function, rule)).hexdigest()

    def invoke(self, name, qualifier=None, payload=None):
        kwargs = {}
        kwargs['FunctionName'] = name
        kwargs['InvocationType'] = 'RequestResponse'
        kwargs['LogType'] = 'Tail'

        if payload is not None:
            kwargs['Payload'] = payload
        if qualifier is not None:
            kwargs['Qualifier'] = qualifier

        return self._lambda.invoke(**kwargs)

    def get_previous_version(self, function, alias):
        ver = self.get_alias(function, previous_alias(alias))
        return ver.get('FunctionVersion')



# -*- coding: utf-8 -*-

from abc import ABCMeta, abstractmethod
from lamvery.client import Client
from lamvery.config import Config
from lamvery.log import get_logger

class BaseAction:

    __metaclass__ = ABCMeta

    _logger = None

    def __init__(self, args):
        self._config = Config(args.conf_file)
        self._dry_run = False

        logger_name = 'lamvery'
        if hasattr(args, 'dry_run'):
            self._dry_run = args.dry_run
            if self._dry_run:
                logger_name = '(Dry run) lamvery'

        self._logger = get_logger(logger_name)

    @abstractmethod
    def action(self):
        raise NotImplementedError

    def get_client(self):
        return Client(
            region=self._config.get_region(),
            profile=self._config.get_profile(),
            dry_run=self._dry_run)

    def _get_diff(self, remote, local, keys):
        diff = {}
        for k in keys:
            r = remote.get(k[0])
            l = local.get(k[1])
            if r == l:
                diff[k[1]] = None
            else:
                diff[k[1]] = (r, l,)
        return diff

    def _print_diff(self, prefix, remote, local, keys):
        diff = self._get_diff(remote, local, keys)
        for k,v in diff.items():
            if v is not None:
                self._logger.warn(
                    '{p} {k}: {r} -> {l}'.format(p=prefix, k=k, r=v[0], l=v[1]))



# -*- coding: utf-8 -*-

import yaml
import os
import uuid
import json
import re
from termcolor import cprint, colored
from collections import OrderedDict
from lamvery.archive import Archive
from lamvery.client import Client
from jinja2 import Environment, FileSystemLoader

def represent_odict(dumper, instance):
     return dumper.represent_mapping(u'tag:yaml.org,2002:map', instance.items())

yaml.add_representer(OrderedDict, represent_odict)
yaml.add_constructor(yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG,
    lambda loader, node: OrderedDict(loader.construct_pairs(node)))

class Config:

    def __init__(self, conf_file):
        self._file = conf_file
        self._template_env = Environment(loader=FileSystemLoader('./', encoding='utf8'))

    def load_conf(self):
        tmpl = self._template_env.get_template(self._file)
        return yaml.load(tmpl.render({'env': os.environ}))

    def load_raw_conf(self):
        txt = open(self._file, 'r').read()
        return yaml.load(self.escape(txt))

    def escape(self, txt):
        txt = txt.replace("'", "''")
        txt = txt.replace(
            self._template_env.variable_start_string,
            "'" + self._template_env.variable_start_string)
        txt = txt.replace(
            self._template_env.block_start_string,
            "'" + self._template_env.block_start_string)
        txt = txt.replace(
            self._template_env.variable_end_string,
            self._template_env.variable_end_string + "'")
        txt = txt.replace(
            self._template_env.block_end_string,
            self._template_env.block_end_string + "'")
        return txt

    def unescape(self, txt):
        txt = txt.replace("''", "'")
        txt = txt.replace(
            "'" + self._template_env.variable_start_string,
            self._template_env.variable_start_string)
        txt = txt.replace(
            "'" + self._template_env.block_start_string,
            self._template_env.block_start_string)
        txt = txt.replace(
            self._template_env.variable_end_string + "'",
            self._template_env.variable_end_string)
        txt = txt.replace(
            self._template_env.block_end_string + "'",
            self._template_env.block_end_string)
        return txt

    def get_configuration(self):
        return self.load_conf().get('configuration')

    def get_secret(self):
        return self.load_conf().get('secret')

    def get_events(self):
        events = self.load_conf().get('events')
        if events is None:
            return []
        return events

    def get_default_alias(self):
        return self.load_conf().get('default_alias')

    def enable_versioning(self):
        return bool(self.load_conf().get('versioning'))

    def generate_lambda_secret(self):
        return {
            'region': self.get_region(),
            'cipher_texts': self.get_secret().get('cipher_texts')
        }

    def get_function_name(self):
        if os.path.exists(self._file):
            return self.get_configuration().get('name')
        else:
            return os.path.basename(os.getcwd())

    def get_function_filename(self):
        runtimes = { "python2.7": "py", "nodejs": "js" }
        ext = runtimes.get(self.get_configuration().get('runtime'), "py")
        return '{}.{}'.format(self.get_configuration().get('handler').split('.')[0], ext)

    def get_archive_name(self):
        return '{}.zip'.format(self.get_function_name())

    def get_region(self):
        if os.path.exists(self._file):
            return self.load_conf().get('region')
        else:
            return None

    def get_profile(self):
        return self.load_conf().get('profile')

    def get_exclude(self):
        exclude = self.load_conf().get('exclude')
        if exclude is None:
            return []
        return exclude

    def get_default(self):
        init_config = OrderedDict()
        init_config['name']          = self.get_function_name()
        init_config['runtime']       = 'python2.7'
        init_config['versioning']    = False
        init_config['default_alias'] = None
        init_config['role']          = 'arn:aws:iam::<account-number>:role/<role>'
        init_config['handler']       = 'lambda_function.lambda_handler'
        init_config['description']   = 'This is a sample lambda function.'
        init_config['timeout']       = 10
        init_config['memory_size']   = 128

        init_secret = OrderedDict()
        init_secret['key_id'] = '<key-id>'
        init_secret['cipher_texts'] = OrderedDict()

        targets = OrderedDict()
        targets = [OrderedDict(
            [('id', '<unique-target-id>',),
            ('input', {'this': [{'is': 'a'}, {'sample': 'input'}]},),
            ('input_path', 'json.path.format',)])]
        event = OrderedDict()
        event['rule']        = 'sample-rule-name'
        event['description'] = 'This is a sample CloudWatchEvent'
        event['schedule']    = 'rate(5 minutes)'
        event['targets']     = targets
        init_events = [event]

        init_yaml = OrderedDict()
        init_yaml['profile'] = None
        init_yaml['region']  = 'us-east-1'
        init_yaml['configuration'] = init_config
        init_yaml['events'] = init_events
        init_yaml['secret'] = init_secret
        init_yaml['exclude'] = ['^{}$'.format(re.escape(self._file))]

        return init_yaml

    def write_default(self):
        self.write(self.get_default())

    def write(self, conf):
        txt = yaml.dump(
            conf,
            default_flow_style=False,
            allow_unicode=True)
        open(self._file, 'w').write(self.unescape(txt))

    def file_exists(self):
        return os.path.exists(self._file)

    def store_secret(self, key, value):
        conf = self.load_raw_conf()
        default = self.get_default()

        if 'secret' not in conf:
            conf['secret'] = default['secret']
        elif 'cipher_texts' not in conf['secret']:
            conf['secret']['cipher_texts'] = {}

        conf['secret']['cipher_texts'][key] = value
        self.write(conf)



#!/usr/bin/env python
import sys
import os

from setuptools import setup, find_packages

long_description = 'Function based deploy and management tool for AWS Lambda.'
if os.path.exists('README.txt'):
    long_description = open('README.txt').read()

setup_options = dict(
    name='lamvery',
    version='0.9.0',
    description='Function based deploy and management tool for AWS Lambda.',
    long_description=long_description,
    author='Masashi Terui',
    author_email='marcy9114+pypi@gmail.com',
    url='https://github.com/marcy-terui/lamvery',
    packages=find_packages(exclude=['tests*', 'lambda_function', 'register']),
    install_requires=open('requirements.txt').read().splitlines(),
    entry_points={
    'console_scripts':
        'lamvery = lamvery.cli:main'
    },
    license="MIT License",
    classifiers=(
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
    ),
    keywords='aws lambda',
)

setup(**setup_options)



# -*- coding: utf-8 -*-

from termcolor import colored
from lamvery.actions.base import BaseAction

class InitAction(BaseAction):

    def __init__(self, args):
        super(InitAction, self).__init__(args)
        self._conf_file = args.conf_file

    def action(self):
        if self._needs_write_conf():
            self._config.write_default()
            self._logger.info(
                'Output initial configuration file to {}'.format(self._conf_file))

    def _needs_write_conf(self):
        ret = True
        if self._config.file_exists():
            y_n = raw_input(
                colored('Overwrite {}? [y/n]: '.format(self._conf_file), 'yellow'))
            if y_n != 'y':
                ret = False
        return ret
